import numpy as np

from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from cftool.misc import register_core
from cftool.misc import shallow_copy_dict
from cftool.misc import get_unique_indices
from cftool.misc import Sampler
from cftool.misc import LoggingMixin

from abc import *
from .misc import *
from ..types import *
from .api import TabularData


class KFold:
    """
    Util class which can perform k-fold data splitting:

    1. X = {x1, x2, ..., xn} -> [X1, X2, ..., Xk]
    2. Xi ∩ Xj = ∅, ∀ i, j = 1,..., K
    3. X1 ∪ X2 ∪ ... ∪ Xk = X

    * Notice that `KFold` does not always hold the principles listed above, because `DataSplitter`
    will ensure that at least one sample of each class will be kept. In this case, when we apply
    `KFold` to an imbalance dataset, `KFold` may slightly violate principle 2. and 3.

    Parameters
    ----------
    k : int, number of folds
    dataset : TabularDataset, dataset which we want to split
    **kwargs : used to initialize `DataSplitter` instance

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular.api import TabularDataset
    >>> from cfdata.tabular.toolkit import KFold
    >>>
    >>> x = np.arange(12).reshape([6, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros(6, np_int_type)
    >>> y[[-1, -2]] = 1
    >>> dataset = TabularDataset.from_xy(x, y, "clf")
    >>> k_fold = KFold(3, dataset)
    >>> for train_fold, test_fold in k_fold:
    >>>     print(np.vstack([train_fold.dataset.x, test_fold.dataset.x]))
    >>>     print(np.vstack([train_fold.dataset.y, test_fold.dataset.y]))

    """

    def __init__(self, k: int, dataset: TabularDataset, **kwargs: Any):
        if k <= 1:
            raise ValueError("k should be larger than 1 in KFold")
        ratio = 1.0 / k
        self.n_list = (k - 1) * [ratio]
        self.splitter = DataSplitter(**kwargs).fit(dataset)
        self._cursor: int
        self._order: np.ndarray
        self.split_results: List[SplitResult]

    def __iter__(self) -> "KFold":
        self.split_results = self.splitter.split_multiple(
            self.n_list,
            return_remained=True,
        )
        self._order = np.random.permutation(len(self.split_results)).tolist()
        self._cursor = 0
        return self

    def __next__(self) -> Tuple[SplitResult, SplitResult]:
        if self._cursor >= len(self._order):
            raise StopIteration
        train_results = self.split_results.copy()
        test_result = train_results.pop(self._order[self._cursor])
        train_result = SplitResult.concat(train_results)
        self._cursor += 1
        return train_result, test_result


class KRandom:
    """
    Util class which can perform k-random data splitting:

    1. X = {x1, x2, ..., xn} -> [X1, X2, ..., Xk]
    2. idx{X1} ≠ idx{X2} ≠ ... ≠ idx{Xk}, where idx{X} = {1, 2, ..., n}
    3. X1 = X2 = ... = Xk = X

    Parameters
    ----------
    k : int, number of folds
    num_test : {int, float}
    * if float and  < 1 : ratio of the test dataset
    * if int   and  > 1 : exact number of test samples
    dataset : TabularDataset, dataset which we want to split
    **kwargs : used to initialize `DataSplitter` instance

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular.api import TabularDataset
    >>> from cfdata.tabular.toolkit import KRandom
    >>>
    >>> x = np.arange(12).reshape([6, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros(6, np_int_type)
    >>> y[[-1, -2]] = 1
    >>> dataset = TabularDataset.from_xy(x, y, "clf")
    >>> k_random = KRandom(3, 2, dataset)
    >>> for train_fold, test_fold in k_random:
    >>>     print(np.vstack([train_fold.dataset.x, test_fold.dataset.x]))
    >>>     print(np.vstack([train_fold.dataset.y, test_fold.dataset.y]))

    """

    def __init__(
        self,
        k: int,
        num_test: Union[int, float],
        dataset: TabularDataset,
        **kwargs: Any,
    ):
        self._cursor: int
        self.k, self.num_test = k, num_test
        self.splitter = DataSplitter(**kwargs).fit(dataset)

    def __iter__(self) -> "KRandom":
        self._cursor = 0
        return self

    def __next__(self) -> Tuple[SplitResult, SplitResult]:
        if self._cursor >= self.k:
            raise StopIteration
        self._cursor += 1
        self.splitter.reset()
        test_result, train_result = self.splitter.split_multiple(
            [self.num_test],
            return_remained=True,
        )
        return train_result, test_result


class KBootstrap:
    """
    Util class which can perform k-random data splitting with bootstrap:

    1. X = {x1, x2, ..., xn} -> [X1, X2, ..., Xk] (Use bootstrap aggregation to collect datasets)
    3. idx{X1} ≠ idx{X2} ≠ ... ≠ idx{Xk}, where idx{X} = {1, 2, ..., n}
    4. X1 = X2 = ... = Xk = X

    * Notice that only some of the special algorithms (e.g. bagging) need `KBootstrap`.

    Parameters
    ----------
    k : int, number of folds
    num_test : {int, float}
    * if float and  < 1 : ratio of the test dataset
    * if int   and  > 1 : exact number of test samples
    dataset : TabularDataset, dataset which we want to split
    **kwargs : used to initialize `DataSplitter` instance

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular.api import TabularDataset
    >>> from cfdata.tabular.toolkit import KBootstrap
    >>>
    >>> x = np.arange(12).reshape([6, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros(6, np_int_type)
    >>> y[[-1, -2]] = 1
    >>> dataset = TabularDataset.from_xy(x, y, "clf")
    >>> k_bootstrap = KBootstrap(3, 2, dataset)
    >>> for train_fold, test_fold in k_bootstrap:
    >>>     print(np.vstack([train_fold.dataset.x, test_fold.dataset.x]))
    >>>     print(np.vstack([train_fold.dataset.y, test_fold.dataset.y]))

    """

    def __init__(
        self,
        k: int,
        num_test: Union[int, float],
        dataset: TabularDataset,
        **kwargs: Any,
    ):
        self._cursor: int
        self.dataset = dataset
        self.num_samples = len(dataset)
        if isinstance(num_test, float):
            num_test = int(round(num_test * self.num_samples))
        self.k, self.num_test = k, num_test
        self.splitter = DataSplitter(**kwargs).fit(dataset)

    def __iter__(self) -> "KBootstrap":
        self._cursor = 0
        return self

    def __next__(self) -> Tuple[SplitResult, SplitResult]:
        if self._cursor >= self.k:
            raise StopIteration
        self._cursor += 1
        self.splitter.reset()
        test_result, train_result = self.splitter.split_multiple(
            [self.num_test],
            return_remained=True,
        )
        tr_indices = train_result.corresponding_indices
        tr_indices = np.random.choice(tr_indices, len(tr_indices))
        tr_set = self.dataset.split_with(tr_indices)
        tr_split = SplitResult(tr_set, tr_indices, None)
        return tr_split, test_result


class ImbalancedSampler(LoggingMixin):
    """
    Util class which can sample imbalance dataset in a balanced way

    Parameters
    ----------
    data : TabularData, data which we want to sample from
    imbalance_threshold : float
    * for binary class cases, if n_pos / n_neg < threshold, we'll treat data as imbalance data
    * for multi  class cases, if n_min_class / n_max_class < threshold, we'll treat data as imbalance data
    shuffle : bool, whether shuffle the returned indices
    sample_method : str, sampling method used in `cftool.misc.Sampler`
    * currently only 'multinomial' is supported
    verbose_level : int, verbose level used in `LoggingMixin`

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular import TabularData
    >>> from cfdata.tabular.toolkit import ImbalancedSampler
    >>> from cftool.misc import get_counter_from_arr
    >>>
    >>> n = 20
    >>> x = np.arange(2 * n).reshape([n, 2])
    >>> # create an imbalance dataset
    >>> y = np.zeros([n, 1], np_int_type)
    >>> y[-1] = [1]
    >>> data = TabularData().read(x, y)
    >>> sampler = ImbalancedSampler(data)
    >>> # Counter({1: 12, 0: 8})
    >>> # This may vary, but will be rather balanced
    >>> # You might notice that positive samples are even more than negative samples!
    >>> print(get_counter_from_arr(y[sampler.get_indices()]))

    """

    def __init__(
        self,
        data: TabularData,
        imbalance_threshold: float = 0.1,
        *,
        shuffle: bool = True,
        aggregation: str = "continuous",
        aggregation_config: Optional[Dict[str, Any]] = None,
        sample_weights: Optional[np.ndarray] = None,
        sample_method: str = "multinomial",
        verbose_imbalance: bool = True,
        verbose_level: int = 2,
    ):
        self.data = data
        self.shuffle = shuffle
        self.imbalance_threshold = imbalance_threshold
        self._sample_imbalance_flag = True
        self._aggregation_name = aggregation
        self._aggregation_config = aggregation_config
        if not data.is_ts:
            self.aggregation = None
            self._num_samples = len(data)
        else:
            if aggregation_config is None:
                aggregation_config = {}
            base = aggregation_dict[aggregation]
            self.aggregation = base(data, aggregation_config, data._verbose_level)
            self._num_samples = len(self.aggregation.indices2id)
        if sample_weights is not None:
            label_counts = None
            self.sample_weights = sample_weights.copy()
            self.sample_weights /= self.sample_weights.sum() + 1e-8
            self._sampler = Sampler(sample_method, self.sample_weights)
        else:
            self.sample_weights = None
            if not self.shuffle or data.is_reg:
                label_counts = self._label_ratios = self._sampler = None
            else:
                label_recognizer = data.recognizers[-1]
                if label_recognizer is None:
                    raise ValueError(
                        "`data` should contain label recognizer "
                        "for `ImbalancedSampler`"
                    )
                label_counter = label_recognizer.counter
                transform_dict = label_recognizer.transform_dict
                new_counter = {transform_dict[k]: v for k, v in label_counter.items()}
                counts_list = [new_counter[k] for k in sorted(new_counter)]
                label_counts = np.array(counts_list, np_float_type)
                self._label_ratios = label_counts / self._num_samples
                max_label_count = label_counts.max()
                if label_counts.min() / max_label_count >= imbalance_threshold:
                    self._sampler = None
                else:
                    processed = data.processed
                    if processed is None:
                        raise ValueError(
                            "`data` should contain `processed` "
                            "for `ImbalancedSampler`"
                        )
                    if not isinstance(processed.y, np.ndarray):
                        raise ValueError(
                            "`data` should contain `processed.y` "
                            "for `ImbalancedSampler`"
                        )
                    labels = processed.y.ravel()
                    sample_weights = np.zeros(self._num_samples, np_float_type)
                    for i, count in enumerate(label_counts):
                        sample_weights[labels == i] = max_label_count / count
                    sample_weights /= sample_weights.sum() + 1e-8
                    self._sampler = Sampler(sample_method, sample_weights)

        self._sample_method = sample_method
        self._verbose_level = verbose_level
        if label_counts is not None and verbose_imbalance:
            if self._sampler is not None:
                self.log_msg(
                    "using imbalanced sampler with "
                    f"label counts = {label_counts.tolist()}",
                    self.info_prefix,
                    2,
                )

    def __len__(self) -> int:
        return self._num_samples

    @property
    def is_imbalance(self) -> bool:
        return self._sampler is not None

    @property
    def sample_imbalance(self) -> bool:
        return self._sample_imbalance_flag

    @property
    def label_ratios(self) -> Optional[np.ndarray]:
        return self._label_ratios

    def switch_imbalance_status(self, flag: bool) -> None:
        self._sample_imbalance_flag = flag

    def get_indices(self) -> np.ndarray:
        if not self.shuffle or not self._sample_imbalance_flag or not self.is_imbalance:
            indices = np.arange(self._num_samples).astype(np_int_type)
        else:
            if self._sampler is None:
                raise ValueError("`_sampler` is not yet generated")
            indices = self._sampler.sample(self._num_samples)
        if self.shuffle:
            np.random.shuffle(indices)
        if self.aggregation is not None:
            indices = self.aggregation.aggregate(indices)
        return indices

    def copy(self) -> "ImbalancedSampler":
        aggregation_config = None
        if self._aggregation_config is not None:
            aggregation_config = shallow_copy_dict(self._aggregation_config)
        sample_weights = None
        if self.sample_weights is not None:
            sample_weights = self.sample_weights.copy()
        return ImbalancedSampler(
            self.data,
            self.imbalance_threshold,
            shuffle=self.shuffle,
            aggregation=self._aggregation_name,
            aggregation_config=aggregation_config,
            sample_weights=sample_weights,
            sample_method=self._sample_method,
            verbose_level=self._verbose_level,
            verbose_imbalance=False,
        )


class DataLoader:
    """
    Util class which can generated batches from `ImbalancedSampler`

    Examples
    ----------
    >>> import numpy as np
    >>>
    >>> from cfdata.types import np_int_type
    >>> from cfdata.tabular import TabularData
    >>> from cfdata.tabular.toolkit import DataLoader, ImbalancedSampler
    >>> from cftool.misc import get_counter_from_arr
    >>>
    >>> n = 20
    >>> x = np.arange(2 * n).reshape([n, 2])
    >>> y = np.zeros([n, 1], np_int_type)
    >>> y[-1] = [1]
    >>> data = TabularData().read(x, y)
    >>> sampler = ImbalancedSampler(data)
    >>> loader = DataLoader(16, sampler)
    >>> y_batches = []
    >>> for x_batch, y_batch in loader:
    >>>     y_batches.append(y_batch)
    >>>     # (16, 1) (16, 1)
    >>>     # (4, 1) (4, 1)
    >>>     print(x_batch.shape, y_batch.shape)
    >>> # Counter({1: 11, 0: 9})
    >>> print(get_counter_from_arr(np.vstack(y_batches).ravel()))

    """

    def __init__(
        self,
        batch_size: int,
        sampler: ImbalancedSampler,
        *,
        return_indices: bool = False,
        label_collator: Optional[Callable[[np.ndarray], np.ndarray]] = None,
        verbose_level: int = 2,
    ):
        self._cursor: int
        self._indices_in_use = None
        self._verbose_level = verbose_level
        self.data = sampler.data
        self.sampler = sampler
        self.return_indices = return_indices
        self._label_collator = label_collator
        self._num_samples = len(sampler)
        self.batch_size = min(self._num_samples, batch_size)

    def __len__(self) -> int:
        n_iter = int(self._num_samples / self.batch_size)
        return n_iter + int(n_iter * self.batch_size < self._num_samples)

    def __iter__(self) -> "DataLoader":
        self._reset()
        return self

    def __next__(self) -> batch_type:
        data_next = self._get_next_batch()
        if self.return_indices:
            (x_batch, y_batch), indices = data_next
        else:
            indices = None
            x_batch, y_batch = data_next
        if self._label_collator is not None:
            y_batch = self._label_collator(y_batch)
        batch = x_batch, y_batch
        if not self.return_indices:
            return batch
        return batch, indices

    @property
    def enabled_sampling(self) -> bool:
        return self.sampler.sample_imbalance

    @enabled_sampling.setter
    def enabled_sampling(self, value: bool) -> None:
        self.sampler.switch_imbalance_status(value)

    def _reset(self) -> None:
        self._cursor = -1
        self._indices_in_use = self.sampler.get_indices()

    def _get_next_batch(self) -> batch_type:
        n_iter, self._cursor = len(self), self._cursor + 1
        if self._cursor == n_iter:
            raise StopIteration
        if self._indices_in_use is None:
            raise ValueError("`_indices_in_use` is not yet generated")
        start = self._cursor * self.batch_size
        end = start + self.batch_size
        indices = self._indices_in_use[start:end]
        batch = self.data[indices]
        if not self.return_indices:
            return batch
        return batch, indices

    def _check_full_batch(self, data_item: data_item_type) -> bool:
        if len(data_item[0]) == self.batch_size:
            return True
        return False

    def copy(self) -> "DataLoader":
        return DataLoader(
            self.batch_size,
            self.sampler.copy(),
            return_indices=self.return_indices,
            label_collator=self._label_collator,
            verbose_level=self._verbose_level,
        )


# time series


aggregation_dict: Dict[str, Type["AggregationBase"]] = {}


class AggregationBase(LoggingMixin, metaclass=ABCMeta):
    def __init__(self, data: TabularData, config: Dict[str, Any], verbose_level: int):
        if not data.is_ts:
            raise ValueError("time series data is required")
        self.data = data
        self.config = config
        self._verbose_level = verbose_level
        self._num_history = config.setdefault("num_history", 1)
        raw = data.raw
        if raw is None:
            raise ValueError("`data` need to contain `raw` for `AggregationBase`")
        if raw.xT is None:
            raise ValueError("`data` need to contain `raw.xT` for `AggregationBase`")
        if data.ts_config is None:
            raise ValueError("`data` need to contain `ts_config` for `AggregationBase`")
        id_column_idx = data.ts_config.id_column_idx
        if id_column_idx is None:
            msg = "`ts_config` need to contain `id_column_idx` for `AggregationBase`"
            raise ValueError(msg)
        id_column = raw.xT[id_column_idx]
        sorted_id_column = [id_column[i] for i in data.ts_sorting_indices]
        unique_indices = get_unique_indices(np.array(sorted_id_column))
        self.indices2id: np.ndarray
        self._unique_id_arr = unique_indices.unique
        self._id2indices = unique_indices.split_indices
        self._initialize()

    @property
    @abstractmethod
    def num_aggregation(self) -> int:
        pass

    @abstractmethod
    def _aggregate_core(self, indices: np.ndarray) -> np.ndarray:
        """indices should be a column vector"""

    def _initialize(self) -> None:
        num_list = list(map(len, self._id2indices))
        self._num_samples_per_id = np.array(num_list, np_int_type)
        self.log_msg("generating valid aggregation info", self.info_prefix, 5)
        valid_mask = self._num_samples_per_id >= self._num_history
        if not valid_mask.any():
            raise ValueError(
                "current settings lead to empty valid dataset, "
                "increasing raw dataset size or decreasing n_history "
                f"(current: {self._num_history}) might help"
            )
        if not valid_mask.all():
            invalid_mask = ~valid_mask
            n_invalid_id = invalid_mask.sum()
            n_invalid_samples = self._num_samples_per_id[invalid_mask].sum()
            self.log_msg(
                f"{n_invalid_id} id (with {n_invalid_samples} samples) "
                f"will be dropped (n_history={self._num_history})",
                self.info_prefix,
                verbose_level=2,
            )
            invalid_ids = self._unique_id_arr[invalid_mask].tolist()
            self.log_msg(
                f"dropped id : {', '.join(map(str, invalid_ids))}",
                self.info_prefix,
                verbose_level=4,
            )
        self._num_samples_per_id_cumsum = np.hstack(
            [[0], np.cumsum(self._num_samples_per_id[:-1])]
        )
        # self._id2indices need to contain 'redundant' indices here because
        # aggregation need to aggregate those 'invalid' samples
        self._id2indices_stack = np.hstack(self._id2indices)
        self.log_msg(
            "generating aggregation attributes",
            self.info_prefix,
            verbose_level=5,
        )
        self._get_id2valid_indices()
        self._inject_valid_samples_info()

    def _inject_valid_samples_info(self) -> None:
        # 'indices' in self.indices2id here doesn't refer to indices of original dataset
        # (e.g. 'indices' in self._id2indices), but refers to indices generated by sampler,
        # so we should only care 'valid' indices here
        self._num_valid_samples_per_id = list(map(len, self._id2valid_indices))
        self._num_valid_samples_per_id_cumsum = np.hstack(
            [[0], np.cumsum(self._num_valid_samples_per_id[:-1])]
        )
        num_int = self._num_valid_samples_per_id_cumsum.astype(np_int_type)
        self._num_valid_samples_per_id_cumsum = num_int
        arange = np.arange(len(self._unique_id_arr))
        self.indices2id = np.repeat(arange, self._num_valid_samples_per_id)
        self._id2valid_indices_stack = np.hstack(self._id2valid_indices)

    def _get_id2valid_indices(self) -> None:
        # TODO : support nan_fill here
        nan_fill = self.config.setdefault("nan_fill", "past")
        nan_ratio = self.config.setdefault("nan_ratio", 0.0)
        self._id2valid_indices = [
            np.array([], np_int_type)
            if len(indices) < self._num_history
            else np.arange(
                cumsum, cumsum + len(indices) - self._num_history + 1
            ).astype(np_int_type)
            for cumsum, indices in zip(
                self._num_samples_per_id_cumsum, self._id2indices
            )
        ]
        self._inject_valid_samples_info()
        processed = self.data.processed
        if processed is None:
            raise ValueError("`processed` is not generated yet")
        x, y = processed.xy
        assert isinstance(x, np.ndarray)
        feature_dim = self.data.processed_dim
        for i, valid_indices in enumerate(self._id2valid_indices):
            cumsum = self._num_valid_samples_per_id_cumsum[i]
            arange = np.arange(cumsum, cumsum + len(valid_indices))
            aggregated_flat_indices = self.aggregate(arange).ravel()
            aggregated_x = x[aggregated_flat_indices]
            shape = [-1, self.num_aggregation, feature_dim]
            aggregated_x = aggregated_x.reshape(shape)
            aggregated_x_nan_mask = np.isnan(aggregated_x)
            if y is None:
                aggregated_y_valid_mask = None
            else:
                aggregated_y = y[self.get_last_indices(aggregated_flat_indices)]
                aggregated_y_valid_mask = ~np.isnan(aggregated_y)
            aggregated_nan_ratio = aggregated_x_nan_mask.mean((1, 2))
            valid_mask = aggregated_nan_ratio <= nan_ratio
            if aggregated_y_valid_mask is not None:
                valid_mask &= aggregated_y_valid_mask.ravel()
            new_valid_indices = valid_indices[valid_mask]
            self._id2valid_indices[i] = new_valid_indices

    def aggregate(self, indices: np.ndarray) -> np.ndarray:
        """
        We've maintained two groups of indices in `_initialize` method:
        * the 'original' indices, which points to indices of original dataset
        * the 'valid' indices, which is 'virtual' should points to indices
        of the 'original' indices

        So we need to translate sampler indices to 'valid' indices, add offsets
        to the 'valid' indices, and then fetch the 'original' indices to fetch
        the corresponding data
        * _aggregate_core method will add offsets for us

        Parameters
        ----------
        indices : np.ndarray, indices come from sampler

        Returns
        -------
        indices : np.ndarray, aggregated 'original' indices

        """
        valid_indices = self._id2valid_indices_stack[indices]
        aggregated_valid_indices_mat = self._aggregate_core(valid_indices[..., None])
        aggregated = self._id2indices_stack[aggregated_valid_indices_mat.ravel()]
        reversed_aggregated = self.data.ts_sorting_indices[aggregated]
        return reversed_aggregated.reshape([-1, self.num_aggregation])

    def get_last_indices(self, aggregated_flat_indices: np.ndarray) -> np.ndarray:
        reshaped = aggregated_flat_indices.reshape([-1, self.num_aggregation])
        return reshaped[..., -1]

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global aggregation_dict
        return register_core(name, aggregation_dict)


@AggregationBase.register("continuous")
class ContinuousAggregation(AggregationBase):
    def _initialize(self) -> None:
        self._history_arange = np.arange(self._num_history)
        super()._initialize()

    @property
    def num_aggregation(self) -> int:
        return self._num_history

    def _aggregate_core(self, indices: np.ndarray) -> np.ndarray:
        return indices + self._history_arange


__all__ = [
    "KFold",
    "KRandom",
    "KBootstrap",
    "ImbalancedSampler",
    "DataLoader",
    "aggregation_dict",
    "AggregationBase",
]



import math
import numpy as np

from typing import Any
from typing import Dict
from cftool.misc import is_numeric


def naive_transform_flat_data_with_dict(
    flat_data: np.ndarray,
    transform_dict: Dict[Any, Any],
    need_truncate: bool,
) -> np.ndarray:
    nan_value = transform_dict.get("nan", 0)
    oob_value = nan_value if need_truncate else 0
    for i, elem in enumerate(flat_data):
        elem = float(elem)
        if math.isnan(elem):
            flat_data[i] = nan_value
        else:
            flat_data[i] = transform_dict.get(elem, oob_value)
    return flat_data


def naive_is_all_numeric(array: np.ndarray) -> bool:
    return all(map(is_numeric, array))


def naive_flat_arr_to_float32(array: np.ndarray) -> np.ndarray:
    return np.asarray(array, np.float32)


__all__ = [
    "naive_transform_flat_data_with_dict",
    "naive_is_all_numeric",
    "naive_flat_arr_to_float32",
]



from typing import Any
from typing import Dict
from typing import Union

try:
    from .cython_utils import *
except ImportError:
    raise


import numpy as np


def _to_list(array: Union[list, np.ndarray]) -> list:
    if isinstance(array, np.ndarray):
        array = array.tolist()
    return array


def c_transform_flat_data_with_dict(
    flat_data: np.ndarray,
    transform_dict: Dict[Any, Any],
    need_truncate: bool,
) -> np.ndarray:
    nan_value = transform_dict.get("nan", 0)
    oob_value = nan_value if need_truncate else 0
    return transform_flat_data_with_dict(  # type: ignore
        flat_data,
        nan_value,
        oob_value,
        transform_dict,
    )


def c_is_all_numeric(array: np.ndarray) -> bool:
    return is_all_numeric(_to_list(array))  # type: ignore


def c_flat_arr_to_float32(array: np.ndarray) -> np.ndarray:
    return flat_arr_to_float32(_to_list(array))  # type: ignore


__all__ = [
    "c_transform_flat_data_with_dict",
    "c_is_all_numeric",
    "c_flat_arr_to_float32",
]



try:
    from .cython_wrappers import (
        c_transform_flat_data_with_dict as transform_flat_data_with_dict,
    )
    from .cython_wrappers import c_flat_arr_to_float32 as flat_arr_to_float32
    from .cython_wrappers import c_is_all_numeric as is_all_numeric
except ImportError:
    from .cython_substitute import (
        naive_transform_flat_data_with_dict as transform_flat_data_with_dict,
    )
    from .cython_substitute import naive_flat_arr_to_float32 as flat_arr_to_float32
    from .cython_substitute import naive_is_all_numeric as is_all_numeric


__all__ = ["transform_flat_data_with_dict", "flat_arr_to_float32", "is_all_numeric"]



import math

import numpy as np
import datatable as dt

from typing import *
from collections import Counter
from cftool.misc import shallow_copy_dict
from cftool.misc import get_counter_from_arr

from ..misc import *
from ...types import *
from .binning import BinningBase
from .binning import BinningError


class Recognizer(DataStructure):
    df: dt.Frame
    info: FeatureInfo
    counter: Counter
    transform_dict: transform_dict_type
    num_unique_bound: Optional[int]

    def __init__(
        self,
        name: str,
        is_np: bool,
        *,
        binning: str = "auto",
        is_label: bool = False,
        is_valid: Optional[bool] = None,
        task_type: task_type_type = TaskTypes.NONE,
        labels: Optional[np.ndarray] = None,
        config: Optional[Dict[str, Any]] = None,
    ):
        self.name = name
        self.is_np = is_np
        self.binning = binning
        self.labels = labels
        self.is_label = is_label
        self.is_valid = is_valid
        self.task_type = parse_task_type(task_type)
        self._init_config(config)

    def __str__(self) -> str:
        return f"Recognizer({self.info.column_type})"

    __repr__ = __str__

    @property
    def num_unique_values(self) -> Union[int, float]:
        if self.info.is_numerical:
            return math.inf
        return len(self._transformed_unique_values)

    def _init_config(self, config: Optional[Dict[str, Any]] = None) -> None:
        if config is None:
            config = {}
        self.config = config
        self._num_thresh = config.setdefault("numerical_threshold", 0.5)
        default_bound = config.setdefault("default_bound", 128)
        self.num_unique_bound = config.setdefault("num_unique_bound", default_bound)

    def _make_invalid_info(
        self,
        msg: Optional[str],
        contains_nan: bool,
        nan_mask: np.ndarray,
    ) -> "Recognizer":
        self.info = FeatureInfo(
            contains_nan,
            None,
            nan_mask=nan_mask,
            is_valid=False,
            msg=msg,
        )
        return self

    def _make_string_info(
        self,
        flat_arr: Optional[np.ndarray],
        is_valid: bool,
        msg: Optional[str],
        unique_values: Optional[np.ndarray] = None,
        sorted_counts: Optional[np.ndarray] = None,
    ) -> FeatureInfo:
        return FeatureInfo(
            False,
            flat_arr,
            msg=msg,
            is_valid=is_valid,
            need_transform=True,
            column_type=ColumnTypes.STRING,
            num_unique_bound=self.num_unique_bound,
            unique_values_sorted_by_counts=unique_values,
            sorted_counts=sorted_counts,
        )

    def _make_dummy_info(
        self,
        np_flat: np.ndarray,
        unique_values: np.ndarray,
        sorted_counts: np.ndarray,
    ) -> FeatureInfo:
        return FeatureInfo(
            None,
            np_flat,
            num_unique_bound=self.num_unique_bound,
            unique_values_sorted_by_counts=unique_values,
            sorted_counts=sorted_counts,
        )

    def _get_transform_dict(
        self,
        info: FeatureInfo,
        check_nan: bool,
        unique_values: Union[List[str], List[float]],
        sorted_counts: np.ndarray,
    ) -> Tuple[transform_dict_type, List[int]]:
        def _core(
            unique_values_: Union[List[str], List[float]],
            indices: Optional[List[int]] = None,
        ) -> transform_dict_type:
            if indices is None:
                indices = list(range(len(unique_values_)))
            iterator = zip(indices, unique_values_)
            td: transform_dict_type = {}
            if not check_nan:
                for i, v in iterator:
                    assert isinstance(v, (str, float))
                    td[v] = i
                return td
            for i, v in iterator:
                assert isinstance(v, float)
                if math.isnan(v):
                    continue
                td[v] = i
            return td

        if self.is_label:
            return _core(unique_values), list(range(len(unique_values)))

        if self.binning != "auto":
            binning_type = self.binning
        else:
            binning_type = "opt" if self.labels is not None else "fuse"
        args = binning_type, self.labels, self.task_type, self.config
        binning = BinningBase.make(*args)
        results = binning.binning(info, sorted_counts, unique_values)
        fused_indices, unique_values, transformed_unique_values = results
        return _core(unique_values, fused_indices), transformed_unique_values

    def _check_string_column(
        self,
        df: dt.Frame,
        np_flat: np.ndarray,
        nan_mask: np.ndarray,
    ) -> Tuple[bool, Optional[FeatureInfo]]:
        if not is_string(df.stype.dtype):
            return False, None
        msg: Optional[str] = None
        np_flat[nan_mask] = "nan"
        self.counter = get_counter_from_arr(np_flat)
        most_common = sorted(
            self.counter.most_common(),
            key=lambda pair: pair[::-1],
            reverse=True,
        )
        unique_values = [elem[0] for elem in most_common]
        sorted_counts = np.array([elem[1] for elem in most_common])
        num_unique_values = len(unique_values)
        if not self.is_valid and num_unique_values == 1:
            msg = (
                f"all values in column {self.name}, which tends to be string column, "
                "are the SAME. It'll be excluded since it might be redundant"
            )
            return True, self._make_string_info(None, False, msg)
        if not self.is_valid and num_unique_values == len(np_flat):
            msg = (
                f"all values in column {self.name}, which tends to be string column, "
                "are DIFFERENT. It'll be excluded since it might be redundant"
            )
            return True, self._make_string_info(None, False, msg)
        unique_ratio = num_unique_values / len(np_flat)
        if not self.is_valid and unique_ratio >= 0.5:
            msg = (
                f"unique values occurred in column {self.name}, which tends to be "
                f"string column, are TOO MANY (ratio={unique_ratio:8.6f}). "
                "It'll be excluded since it might be redundant"
            )
            return True, self._make_string_info(None, False, msg)
        unique_values_arr = np.array(unique_values)
        dummy_info = self._make_dummy_info(np_flat, unique_values_arr, sorted_counts)
        args = dummy_info, False, unique_values, sorted_counts
        try:
            pack = self._get_transform_dict(*args)
        except BinningError:
            self.binning = "fuse"
            pack = self._get_transform_dict(*args)
        except Exception as err:
            msg = (
                f"column {self.name}, which tends to be string column, "
                f"cannot be transformed ({err})."
            )
            if not self.is_valid:
                msg = f"{msg} It'll be excluded since it might be redundant"
                return True, self._make_string_info(None, False, msg)
            self.binning = "fuse"
            pack = self._get_transform_dict(*args)
        self.transform_dict, self._transformed_unique_values = pack
        return True, self._make_string_info(
            np_flat,
            True,
            msg,
            unique_values_arr,
            sorted_counts,
        )

    def _check_exclude_categorical(
        self,
        num_samples: int,
        num_unique_values: int,
    ) -> Tuple[str, Optional[str]]:
        if not self.is_valid and num_unique_values == 1:
            msg = (
                f"all values in column {self.name}, which tends to be "
                "categorical column, are the SAME. It'll be excluded "
                "since it might be redundant"
            )
            return "exclude", msg
        if not self.is_valid and num_samples == num_unique_values:
            msg = (
                f"all values in column {self.name}, which tends to be "
                "categorical column, are DIFFERENT. It'll be excluded "
                "since it might be redundant"
            )
            return "exclude", msg
        if num_unique_values >= self._num_thresh * num_samples:
            msg = (
                f"TOO MANY unique values occurred in column {self.name} "
                f"({num_unique_values:^12d}) which tends to be categorical column, "
                "it'll cast to numerical column to save memory "
                "and possibly for better performance"
            )
            return "numerical", msg
        return "keep", None

    def _generate_categorical_transform_dict(self) -> None:
        unique_values = self.info.unique_values_sorted_by_counts
        sorted_counts = self.info.sorted_counts
        assert unique_values is not None and sorted_counts is not None
        values = list(map(float, unique_values.tolist()))
        pack = self._get_transform_dict(self.info, True, values, sorted_counts)
        transform_dict, self._transformed_unique_values = pack
        if self.info.contains_nan:
            num_transformed_unique = len(self._transformed_unique_values)
            transform_dict["nan"] = num_transformed_unique
            self._transformed_unique_values.append(num_transformed_unique)
        self.transform_dict = transform_dict

    def fit(self, df: dt.Frame, *, is_preset: bool) -> "Recognizer":
        if self.is_label and is_preset:
            raise ValueError("`is_preset` should always be False when `is_label`")
        msg: Optional[str]
        dtype = df.stype.dtype
        # check name
        if len(df.names) == 1:
            if df.names[0] != self.name:
                raise ValueError(
                    f"df.name ({df.names[0]}) is not identical with "
                    f"Recognizer.name ({self.name})"
                )
        else:
            for name in df.names:
                if not name.startswith(self.name):
                    raise ValueError(
                        f"one name ({name}) in df.names does not start with "
                        f"Recognizer.name ({self.name})"
                    )
        # check valid
        if self.is_valid is False:
            self.info = FeatureInfo(
                None,
                None,
                False,
                msg=f"current column ({self.name}) is forced to be invalid",
            )
            return self
        # check string
        df_data = df.to_numpy()
        if isinstance(df_data, np.ma.core.MaskedArray):
            np_flat = df_data.data.ravel()
            nan_mask = df_data.mask.ravel()
        else:
            np_flat = df_data.ravel()
            nan_mask = np.zeros_like(np_flat, dtype=np.bool)
        is_string_, info = self._check_string_column(df, np_flat, nan_mask)
        if is_string_:
            assert info is not None
            self.info = info
            return self
        # check bool
        if is_bool(dtype):
            dtype = np.int64
            np_flat = np_flat.astype(dtype)
        # gather info
        valid_mask = ~nan_mask
        np_flat_valid = np_flat[valid_mask]
        np_flat_valid_int = np_flat_valid.astype(np_int_type)
        num_samples, num_valid_samples = map(len, [np_flat, np_flat_valid])
        num_nan = num_samples - num_valid_samples
        contains_nan = num_nan != 0
        if not contains_nan:
            nan_mask = None
        # check whether all nan or not
        if num_valid_samples == 0:
            if self.is_label:
                raise ValueError("label column is detected as all-invalid")
            if self.is_valid:
                np_flat = np.zeros_like(np_flat)
                self.info = FeatureInfo(
                    contains_nan,
                    np_flat,
                    nan_mask=np.zeros_like(nan_mask),
                )
                return self
            msg = (
                f"all values in column {self.name}, which tends to be "
                "categorical column, are NaN. It'll be excluded since "
                "it might be redundant"
            )
            return self._make_invalid_info(msg, contains_nan, nan_mask)
        # check whether it's a numerical column
        like_int = np.allclose(np_flat_valid, np_flat_valid_int)
        not_int = is_float(dtype) and (is_preset or not like_int)
        is_reg_label = self.is_label and self.task_type.is_reg
        if not_int or is_reg_label:
            msg, is_valid = None, True
            if (
                self.is_valid is None
                and np_flat_valid.max() - np_flat_valid.min() <= 1e-8
            ):
                is_valid = False
                msg = (
                    f"all values in column {self.name}, which tends to be "
                    "numerical column, are ALL CLOSE. It'll be excluded since "
                    "it might be redundant"
                )
            self.info = FeatureInfo(
                contains_nan,
                np_flat.astype(np_float_type),
                nan_mask=nan_mask,
                is_valid=is_valid,
                msg=msg,
            )
            return self
        # deal with categorical column
        min_feat = np_flat_valid_int.min()
        np_flat_valid_int -= min_feat
        max_feature_value = np_flat_valid_int.max()
        if max_feature_value > 1e6:
            unique_values, counts = np.unique(np_flat_valid_int, return_counts=True)
            num_unique_values = len(unique_values)
            need_transform = unique_values[-1] != num_unique_values - 1
        else:
            counts = np.bincount(np_flat_valid_int)
            unique_values = np.nonzero(counts)[0]
            num_unique_values = len(unique_values)
            need_transform = num_unique_values != len(counts)
            counts = counts[unique_values]
        need_transform = min_feat != 0 or need_transform
        status, msg = self._check_exclude_categorical(num_samples, num_unique_values)
        if not self.is_label and status != "keep":
            if status == "numerical":
                self.info = FeatureInfo(
                    contains_nan,
                    np_flat,
                    nan_mask=nan_mask,
                    msg=msg,
                )
                return self
            return self._make_invalid_info(msg, contains_nan, nan_mask)
        sorted_indices = np.argsort(counts)[::-1]
        unique_values = unique_values[sorted_indices] + min_feat
        sorted_counts = counts[sorted_indices].astype(np_float_type)
        counter_dict = dict(zip(unique_values, counts))
        if contains_nan:
            unique_values = np.append(unique_values, [float("nan")])
            sorted_counts = np.append(sorted_counts, [float(num_nan)])
            counter_dict["nan"] = num_nan
            # we need to go back to float32 to inject nan so we can
            #  correctly transform categorical columns into 'ordered' indices later on
            np_flat = np_flat.copy().astype(np.float32)
            np_flat[nan_mask] = np.nan
        need_transform = need_transform or contains_nan
        self.counter = Counter(counter_dict)
        self.info = FeatureInfo(
            contains_nan,
            np_flat,
            nan_mask=nan_mask,
            need_transform=need_transform,
            column_type=ColumnTypes.CATEGORICAL,
            num_unique_bound=self.num_unique_bound,
            unique_values_sorted_by_counts=unique_values,
            sorted_counts=sorted_counts,
        )
        try:
            self._generate_categorical_transform_dict()
        except BinningError:
            self.binning = "fuse"
            self._generate_categorical_transform_dict()
        except Exception as err:
            msg = (
                f"column {self.name}, which tends to be categorical column, "
                f"cannot be transformed ({err})."
            )
            if not self.is_valid:
                msg = f"{msg} It'll be excluded since it might be redundant"
                return self._make_invalid_info(msg, contains_nan, nan_mask)
            self.binning = "fuse"
            self._generate_categorical_transform_dict()
        return self

    def dumps_(self) -> Any:
        instance_dict = shallow_copy_dict(self.__dict__)
        instance_dict["_info"] = FeatureInfo(
            self.info.contains_nan,
            None,
            self.info.is_valid,
            None,
            self.info.need_transform,
            self.info.column_type,
            self.info.num_unique_bound,
            self.info.unique_values_sorted_by_counts,
            self.info.sorted_counts,
            self.info.msg,
        )
        return instance_dict

    @classmethod
    def loads(cls, instance_dict: Dict[str, Any], **kwargs: Any) -> "Recognizer":
        recognizer = cls("", False)
        recognizer.__dict__.update(instance_dict)
        return recognizer


__all__ = ["Recognizer"]



import numpy
from Cython.Build import cythonize
from setuptools import setup, find_packages, Extension

VERSION = "0.2.5"

DESCRIPTION = "Data processing module implemented with numpy"
with open("README.md") as f:
    LONG_DESCRIPTION = f.read()

setup(
    name="carefree-data",
    version=VERSION,
    packages=find_packages(exclude=("tests",)),
    install_requires=[
        "optbinning",
        "datatable==0.11.1",
        "carefree-toolkit>=0.2.0",
        "dill",
        "future",
        "psutil",
        "pillow",
        "cython>=0.29.12",
        "numpy>=1.19.2",
        "scipy>=1.2.1",
        "scikit-learn>=0.20.3",
        "matplotlib>=3.0.3",
    ],
    ext_modules=cythonize(
        Extension(
            "cfdata.misc.c.cython_utils",
            sources=["cfdata/misc/c/cython_utils.pyx"],
            language="c",
            include_dirs=[numpy.get_include(), "cfdata/misc/c"],
            library_dirs=[],
            libraries=[],
            extra_compile_args=[],
            extra_link_args=[],
        )
    ),
    package_data={"cfdata.misc.c": ["cython_utils.pyx"]},
    author="carefree0910",
    author_email="syameimaru_kurumi@pku.edu.cn",
    url="https://github.com/carefree0910/carefree-data",
    download_url=f"https://github.com/carefree0910/carefree-data/archive/v{VERSION}.tar.gz",
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords="python numpy data-science",
)
